#include <vector>
#include <stdexcept>

#include <opencv2/core.hpp>

using namespace std;

#include "2702math.h"
#include "2702rigid.h"


using namespace Rebels;
using namespace cv;

Rebels::Point3d Centroid( vector<Rebels::Point3d> const &vPos )
{
	Rebels::Point3d p3dCent;

	for( Rebels::Point3d const & p3d : vPos ) {
		p3dCent += p3d;
	}

	p3dCent /= vPos.size();
	return p3dCent;
}

#if 0
void Q3dRigid( vector<Rebels::Point3d> const &vRefPos, vector<Rebels::Point3d> const &vMeasPos, Rebels::Quaternion &qRot, Rebels::Point3d &p3dTrans )
{
	if( vRefPos.size() != vMeasPos.size() )
		throw std::runtime_error( "fQ3dRigid: input point vectors differ in length" );

	Rebels::Point3d p3dRefCent = Centroid( vRefPos );
	Rebels::Point3d p3dMeasCent = Centroid( vMeasPos );

	float fSxx = 0.f;
	float fSxy = 0.f;
	float fSxz = 0.f;
	float fSyx = 0.f;
	float fSyy = 0.f;
	float fSyz = 0.f;
	float fSzx = 0.f;
	float fSzy = 0.f;
	float fSzz = 0.f;

	for( int i = 0; i < vRefPos.size(); i++ ) {
		Rebels::Point3d p3dRef = vRefPos[i] - p3dRefCent;
		Rebels::Point3d p3dMeas = vMeasPos[i] - p3dMeasCent;

		fSxx += p3dRef.x * p3dRef.x;
		fSxy += p3dRef.x * p3dRef.y;
		fSxz += p3dRef.x * p3dRef.z;

		fSyx += p3dRef.y * p3dMeas.x;
		fSyy += p3dRef.y * p3dMeas.y;
		fSyz += p3dRef.y * p3dMeas.z;

		fSzx += p3dRef.z * p3dMeas.x;
		fSzy += p3dRef.z * p3dMeas.y;
		fSzz += p3dRef.z * p3dMeas.z;
	}

	Mat N( 4, 4, CV_32F );
	N = 0.F;

	N.at(0,0) = fSxx + fSyy + fSzz;
	N.at(0,1) = fSyz - fSzy;
	N.at(0,2) = fSzx - fSxz;
	N.at(0,3) = fSxy - fSyx;

	N.at(1,0) = N.at(0,1);
	N.at(1,1) = fSxx - fSyy - fSzz;
	N.at(1,2) = fSxy + fSyx;
	N.at(1,3) = fSzx + fSxz;

	N.at(2,0) = N.at(0,2);
	N.at(2,1) = N.at(1,2);
	N.at(2,2) = -fSxx + fSyy - fSzz;
	N.at(2,3) = fSyz + fSzy;

	N.at(3,0) = N.at(0,3);
	N.at(3,1) = N.at(1,3);
	N.at(3,2) = N.at(2,3);
	N.at(3,3) = -fSxx -fSyy + fSzz;

	
}
cv::Vec3d
CalculateMean(const cv::Mat_<cv::Vec3d> &points)
{
    cv::Mat_<cv::Vec3d> result;
    cv::reduce(points, result, 0, CV_REDUCE_AVG);
    return result(0, 0);
}
#endif

cv::Mat_<double>
FindRigidTransform(const cv::Mat_<cv::Vec3d> &points1, const cv::Mat_<cv::Vec3d> points2)
{
    /* Calculate centroids. */
    cv::Vec3d t1 = -CalculateMean(points1);
    cv::Vec3d t2 = -CalculateMean(points2);

    cv::Mat_<double> T1 = cv::Mat_<double>::eye(4, 4);
    T1(0, 3) = t1[0];
    T1(1, 3) = t1[1];
    T1(2, 3) = t1[2];

    cv::Mat_<double> T2 = cv::Mat_<double>::eye(4, 4);
    T2(0, 3) = -t2[0];
    T2(1, 3) = -t2[1];
    T2(2, 3) = -t2[2];

    /* Calculate covariance matrix for input points. Also calculate RMS deviation from centroid
     * which is used for scale calculation.
     */
    cv::Mat_<double> C(3, 3, 0.0);
    double p1Rms = 0, p2Rms = 0;
    for (int ptIdx = 0; ptIdx < points1.rows; ptIdx++) {
        cv::Vec3d p1 = points1(ptIdx, 0) + t1;
        cv::Vec3d p2 = points2(ptIdx, 0) + t2;
        p1Rms += p1.dot(p1);
        p2Rms += p2.dot(p2);
        for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
                C(i, j) += p2[i] * p1[j];
            }
        }
    }

    cv::Mat_<double> u, s, vh;
    cv::SVD::compute(C, s, u, vh);

    cv::Mat_<double> R = u * vh;

    if (cv::determinant(R) < 0) {
        R -= u.col(2) * (vh.row(2) * 2.0);
    }

    double scale = sqrt(p2Rms / p1Rms);
    R *= scale;

    cv::Mat_<double> M = cv::Mat_<double>::eye(4, 4);
    R.copyTo(M.colRange(0, 3).rowRange(0, 3));

    cv::Mat_<double> result = T2 * M * T1;
    result /= result(3, 3);

    return result.rowRange(0, 3);
}
