#ifndef DUALBPICAM_H
#define DUALBPICAM_H

// maximum # of managed exposure times.
#define MAX_EXPOSURES	2
#define NUM_WORKERS	4

#define LEFT_SENSOR		0
#define RIGHT_SENSOR	1

#define NUM_APRIL_TAGS	8

struct AprilTagInfo
{
	Rebels::Point2d 	m_grp2dCorners[2][4];
	bool				m_grbValid2d[2];

public:
	AprilTagInfo()
	{
	}
	~AprilTagInfo()
	{
	}

	bool &bValid2d( int nSensor )
	{
		return m_grbValid2d[nSensor];
	}
	Point2d &Corner( int nSensor, int nCorner )
	{
		return m_grp2dCorners[nSensor][nCorner];
	}
	void ClearFrame()
	{
		for( int i = 0; i < 2; i++ ) {
			m_grbValid2d[i] = false;
		}
	}
};

class BPISensor : public v4lcam
{
public:
	BPISensor( int fd );
	virtual ~BPISensor();

	virtual bool bStart();
	virtual bool bStop();
};

class DualBPICam
{
protected:
	v4lcam		m_vcLeft;
	v4lcam		m_vcRight;

	std::mutex				m_mtxQueue;
	std::condition_variable	m_cvQueue;

	std::thread	m_thCamera;
	std::thread	m_th3DProc;

	bool		m_b3DThreadWaiting = false;
	bool		m_bShuttingDown = false;

	v4l2_buffer m_grLeftBuffers[2];
	v4l2_buffer m_grRightBuffers[2];

	int			m_nWorkBuff = -1;

	FILE		*m_fpRaw = 0;			// for logging raw data

	AprilTagInfo	m_grTagInfo[ NUM_APRIL_TAGS ];
	
	StereoCam	m_sc;

	nt::DoubleArrayPublisher		m_dapXYR;

	void vInitSensor( v4lcam &vc );
	void cameraThreadFn();
	void calc3DThread();

	int nDetectAprilTag( uint8_t const **ppu8Img, int i );

	bool bQueueFrame( int nBuffer );
	int nWaitForFrame();
#if defined(MJPEG)
	cs::CvSource m_grcvsOutput[2];
#endif
	
public:
	DualBPICam( char const *pszDev1, char const *pszDev2, char const *pszCamFile );
	DualBPICam( int fd1, int fd2 );
	~DualBPICam();

	void SetRawLog( char const *psz );
};
#endif

