#include <string>
#include <stdexcept>
#include <iostream>
#include <fstream>

using namespace std;

#include <opencv2/imgcodecs.hpp>
#include "VisLog.h"


VisionLog::VisionLog( char const *psz ) 
{
	m_if.open(psz);
	if( !m_if.is_open() )
		throw std::runtime_error( string( "Unable to open input file" ) );
}

VisionLog::~VisionLog() 
{
}

int32_t VisionLog::i32ReadVal()
{
	string str;
	if( !getline( m_if, str ) )
		throw runtime_error( string( "VisionLog::u32ReadVal() - unexpected end of file?" ) );

	return atoi( str.c_str() );
}

// returns sequence #
int32_t VisionLog::i32ReadNext( cv::Mat &img, int &nExp )
{
	if( !m_if.is_open() )
		throw runtime_error( string( "VsionLog::bReadNext() - file is not open" ) );

	string str;
	if( !getline( m_if, str ) )
		return -1;

	if( !str.compare( "<<FRMPNG>>" )) {
		int nSeq = i32ReadVal();
		nExp = i32ReadVal();
		int nSize = i32ReadVal();

		uint8_t gru8Buff[nSize];
		if( !m_if.read( (char *)gru8Buff, nSize ) )
			throw runtime_error( string( "VisionLog::bReadNext() - failed reading frame" ) );

		cv::Mat matImg;
		matImg = cv::imdecode(cv::Mat(1, nSize, CV_8UC1, gru8Buff), cv::IMREAD_UNCHANGED);
		img = matImg;

		return nSeq;
	}
	else if( !str.compare( "<<FRM>>" ) ) {
		int nSeq = i32ReadVal();
		nExp = i32ReadVal();
		int nRows = i32ReadVal();
		int nCols = i32ReadVal();

		cv::Mat matIn( nRows, nCols, CV_8UC1 );

		if( !m_if.read( (char *)matIn.ptr(), nRows * nCols ) ) 
			throw runtime_error( string( "VisionLog::bReadNext() - failed reading frame" ) );

		img = matIn;
		return nSeq;
	}
	else {
		if( m_if.eof() )
			return -1;
		throw runtime_error( string( "VisionLog::bReadNext() - <<FRM>> not found" ) );
	}

	return -1;
}
