#include <sys/time.h>

#include <iostream>
#include <fstream>


using namespace std;

#include <opencv2/core/core.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/calib3d.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include <frc/apriltag/AprilTagPoseEstimator.h>
#include <frc/apriltag/AprilTagDetector.h>
#include <frc/apriltag/AprilTagDetection.h>
#include <frc/geometry/Transform3d.h>
#include <frc/geometry/Pose3d.h>

using namespace cv;
using namespace frc;

#include "field.h"
#include "2702camera.h"
#include "2702math.h"
#include "Bayer2RGB.h"
#include <VisLog.h>

#include "PoseEstimator.h"

static FieldXfrms myField;

Rebels::CamParm cp;		// Camera parameters

cv::Scalar grColors[] = {
	cv::Scalar( 255, 255, 255 ),
	cv::Scalar( 255,   0,   0 ),
	cv::Scalar(   0, 255,   0 ),
	cv::Scalar(   0,   0, 255 )
};

Mat Bayer2Grey( Mat const &imgBayer )
{
	Mat img( imgBayer.rows / 2, imgBayer.cols / 2, CV_8UC1 );
	for( int r = 0; r < img.rows; r++ ) {
		uint8_t const *pu8Src1 = imgBayer.ptr(r * 2);
		uint8_t const *pu8Src2 = imgBayer.ptr(r * 2 + 1);
		uint8_t *pu8Dst = img.ptr(r);
		for( int c = 0; c < img.cols; c++ ) {
			uint32_t u32Sum = pu8Src1[0] + pu8Src1[1] + pu8Src2[0] + pu8Src2[1];
			*pu8Dst++ = u32Sum / 4;
			pu8Src1+=2;
			pu8Src2+=2;
		}
	}

	return img;
}

void Undistort( float &fU, float &fV )
{
	float fA = fU;
	float fB = fV;
	fU = (fU - cp.pinhole(0,2)) / cp.pinhole(0,0);
	fV = (fV - cp.pinhole(1,2)) / cp.pinhole(1,1);

	float fUFU;
	float fUFV;
	cp.Undistort( fU, fV, fUFU, fUFV );

	fU = fUFU * cp.pinhole(0,0) + cp.pinhole(0,2);
	fV = fUFV * cp.pinhole(1,1) + cp.pinhole(1,2);

//	printf( "%8.2f -> %8.2f, %8.2f -> %8.2f\n", fA, fU, fB, fV );
}

double dCalcBlurr( cv::Mat &img )
{
	cv::Mat laplacianImage;

	Laplacian(img, laplacianImage, CV_64F);
	Scalar mean, stddev; // 0:1st channel, 1:2nd channel and 2:3rd channel
	meanStdDev(laplacianImage, mean, stddev, Mat());
	double variance = stddev.val[0] * stddev.val[0];

	return variance;
}

frc::Quaternion AxisAngle2Quat( cv::Mat &rvec )
{
	double angle = norm(rvec);
    double angle_2 = angle / 2;
	double sin_a2 = sin(angle_2);

	double daX = (rvec.at<double>(0) / angle) * sin_a2;
	double daY = (rvec.at<double>(1) / angle) * sin_a2;
	double daZ = (rvec.at<double>(2) / angle) * sin_a2;

	return frc::Quaternion( cos(angle_2), daX, daY, daZ );
    //Matx41d q(axis(0) * sin(angle_2), axis(1) * sin(angle_2), axis(2) * sin(angle_2), cos(angle_2));

    //return q;
}

frc::Transform3d CalcXfrm( vector<cv::Point2f> &v2dPts, vector<cv::Point3f> &v3dPts )
{
#if 0
	std::cout << "There are " << v2dPts.size() << " imagePoints and " << v3dPts.size() << " objectPoints." << std::endl;
	for( int i = 0; i < v2dPts.size(); i++ ) {
		printf( "  %6.2f %6.2f -> %8.3f %8.3f %8.3f\n", v2dPts[i].x, v2dPts[i].y, v3dPts[i].x, v3dPts[i].y, v3dPts[i].z );
	}
#endif
	cv::Mat cameraMatrix(3,3,cv::DataType<double>::type);
	for( int i = 0; i < 3; i++ ) {
		for( int j = 0; j < 3; j++ ) {
			cameraMatrix.at<double>(i,j) = cp.pinhole(i,j);
		}
	}
 
//	std::cout << "Initial cameraMatrix: " << cameraMatrix << std::endl;
 
	cv::Mat distCoeffs(4,1,cv::DataType<double>::type);
	for( int i = 0; i < 4; i++ ) {
  		distCoeffs.at<double>(i) = cp.dist(i);
	}
 
 //	std::cout << "Distortion: " << distCoeffs << std::endl;

	cv::Mat rvec(3,1,cv::DataType<double>::type);
	cv::Mat tvec(3,1,cv::DataType<double>::type);

	tvec.at<double>(3) = 18.f;
	if( !cv::solvePnP(v3dPts, v2dPts, cameraMatrix, distCoeffs, rvec, tvec )) {
		units::length::inch_t inM1(-1.F);
		units::angle::radian_t rdM1(-1.F);
		return Transform3d( Translation3d(inM1, inM1, inM1), Rotation3d( rdM1, rdM1, rdM1 ) );
	}
	//cv::solvePnP(v3dPts, v2dPts, cameraMatrix, distCoeffs, rvec, tvec, SOLVEPNP_IPPE);
	//cv::solvePnP(v3dPts, v2dPts, cameraMatrix, distCoeffs, rvec, tvec, SOLVEPNP_SQPNP);
	cv::solvePnPRefineLM(v3dPts, v2dPts, cameraMatrix, distCoeffs, rvec, tvec );

cv::Mat tvec2(3,1,cv::DataType<double>::type);
cv::Mat R;
cv::Rodrigues(rvec, R); // R is 3x3

R = R.t();  // rotation of inverse
tvec2 = -R * tvec; // translation of inverse

//	std::cout << "rvec:  " << rvec << std::endl;
//	std::cout << "tvec:  " << tvec << std::endl;
//	std::cout << "tvec2: " << tvec2 << std::endl;

	Translation3d t3d( units::length::inch_t(tvec.at<double>(0)),
					   units::length::inch_t(tvec.at<double>(1)),
					   units::length::inch_t(tvec.at<double>(2)) );
	Rotation3d r3d( AxisAngle2Quat( rvec ) );
	return frc::Transform3d( t3d, r3d );
}

Mat Bayer2Green( Mat &imgBayer )
{
	Mat grn( imgBayer.rows, imgBayer.cols/2, CV_8UC1 );
	for( int r = 0; r < imgBayer.rows; r++ ) {
		uint8_t const *pSrc = imgBayer.ptr( r ) + (r & 1);
		uint8_t *pDst = grn.ptr( r );
		for( int c = 0; c < grn.cols; c++, pDst++, pSrc+=2) {
			*pDst = *pSrc;
		}
	}

	return grn;
}

void printHeadingData( char const *psz, Transform3d &xfm )
{
	double fX = xfm.X().value();
	double fZ = xfm.Z().value();

	float fHeading = atan(fX / fZ);

	float fDist = sqrt(fX * fX + fZ * fZ);
	printf( "%s  Heading: %8.2lf  Dist: %8.4lf\n", psz, fHeading * 180. / PI, fDist );
}

void drawHeadingData( int nTagId, Mat &img, Transform3d &xfm )
{
	float fZScale = img.cols / 54.F;
	float fXScale = img.rows / 54.F;

	Point ptRobot( 0, 27.F * fXScale );
	Point ptTag( xfm.Z().value() * 3.28084 * fZScale, (27.F + xfm.X().value() * 3.2804F) * fXScale );
	
	line( img, ptRobot, ptTag, grColors[nTagId&3], 2, LINE_AA );

	float fTheta = xfm.Rotation().Y().value();

	float fSinTheta = sin(fTheta);
	float fCosTheta = cos(fTheta);

	Point ptEnd((0.F * fCosTheta - 1.F * fSinTheta) * fZScale + ptTag.x, (0.F * fSinTheta + 1.F * fCosTheta) * fXScale + ptTag.y );

	line( img, ptTag, ptEnd, grColors[nTagId&3], 2, LINE_AA );
	
}

int main( int argc, char **argv )
{
	if( argc != 2 ) {
		cerr << "Usage: vistune camlog\n";
		return 1;
	}

	cv::Mat matField = imread( "frcfield.png" );
	if( matField.empty() ) {
		cerr << "Unable to load frcfield.png\n";
		return 1;
	}

	vector<cv::Point3f> vTagPts;
	vTagPts.push_back( Point3f( -3.0f, -3.0f, 0.f ) );
	vTagPts.push_back( Point3f(  3.0f, -3.0f, 0.f ) );
	vTagPts.push_back( Point3f(  3.0f,  3.0f, 0.f ) );
	vTagPts.push_back( Point3f( -3.0f,  3.0f, 0.f ) );

	if( !cp.Load("cam.cam") ) {
		cerr << "Unable to load cam.cam\n";
		return 1;
	}
	
	units::length::meter_t m0( 0.F );
	units::angle::radian_t r0( 0.F );
	units::angle::radian_t rP2( PI/2 );

//	Rotation3d r3dTot = Rotation3d( r0, r0, rP2 ) + Rotation3d( rP2, r0, r0 );
	Rotation3d r3dTot = Rotation3d( r0, r0, r0 );

	Transform3d m_x3CamToRobot = Transform3d( Translation3d( m0, m0, m0 ), r3dTot );
	Pose3d m_p3d = Pose3d( Translation3d( units::length::inch_t(0.0), units::length::inch_t(0.0), units::length::inch_t(0.0) ),
					Rotation3d( units::angle::radian_t(0.0), units::angle::radian_t(0.0), units::angle::radian_t(0.0)) );

	AprilTagDetector atd;
	atd.AddFamily("tag16h5", 0);
	AprilTagPoseEstimator::Config atPEstConfig = {
		.tagSize = units::length::inch_t(6.0),
		.fx = cp.pinhole( 0, 0 ),
		.fy = cp.pinhole( 1, 1 ),
		.cx = cp.pinhole( 0, 2 ),
		.cy = cp.pinhole( 1, 2 ),
	};

	AprilTagPoseEstimator atPEst	= AprilTagPoseEstimator( atPEstConfig );

	VisionLog vl(argv[1]);
	namedWindow( "Camera",WINDOW_AUTOSIZE);
	namedWindow( "Field",WINDOW_AUTOSIZE);
	namedWindow( "Tag", WINDOW_AUTOSIZE);
//	namedWindow( "Green", WINDOW_AUTOSIZE);
	namedWindow( "Heading", WINDOW_AUTOSIZE);

	int nSeq;
	int nExp;
	Mat img;
	Mat clr(480,736,CV_8UC3);

	while( (nSeq = vl.i32ReadNext( img, nExp)) > 0 ) {
		cv::Mat imgHeading( 540, 540, CV_8UC3, Scalar(0,0,0) );
		Bayer2RGB( clr, img );
//		cv::cvtColor(clr, img, COLOR_BGR2GRAY);
		cv::Mat matF1 = matField.clone();
//		img = Bayer2Grey( img );
//		clr = img.clone();
		cout << "===============================================================================================\n";
//		Mat grn = Bayer2Green( img );
//		imshow( "Green", grn );
#if 0
		for( int r = 0; r < img.rows; r++ ) {
			uint8_t *p = img.ptr(r);
			
			uint8_t u8Prev = *p++;
			for( int c = 1; c < img.cols - 1; c++ ) {
				uint8_t u8Cur = *p;
				*p = (u8Prev + u8Cur + p[1]) / 3;
				p++;
				u8Prev = u8Cur;
			}
		}
#endif
#if 0
		Mat grey;
		cv::cvtColor( clr, grey, COLOR_BGR2GRAY );
		AprilTagDetector::Results detections = atd.Detect(img.cols, img.rows, (uint8_t *)grey.ptr() );
#else
		AprilTagDetector::Results detections = atd.Detect(img.cols, img.rows, (uint8_t *)img.ptr() );
#endif
		int nDL = 8;

		vector<cv::Point3f> vFieldPts;
		vector<cv::Point2f> vCamCoords;

		for (AprilTagDetection const *detection : detections ) {
			// filter by property
			// run pose estimator
			//Transform3d pose;
			Pose3d pose;

			if (detection->GetId() >= 1 && detection->GetId() <= 8) {
				struct timespec tsCur;
				struct timeval tvCur;
				struct timeval tvDelta;

				double grUOrg[8];
				double grU[8];

				for( int i = 0; i < 4; i++ ) {
					auto p = detection->GetCorner(i);
					float fU = p.x;
					float fV = p.y;
					grUOrg[i * 2]     = fU;
					grUOrg[i * 2 + 1] = fV;
					Undistort( fU, fV );
					grU[i * 2]     = fU;
					grU[i * 2 + 1] = fV;
				}
				Transform3d x3dTag = atPEst.Estimate(*detection);
				printf( "1) %8.3f %8.3f %8.3f\n", x3dTag.X().value(), x3dTag.Y().value(), x3dTag.Z().value() );
#if 1
				printHeadingData( "Uncorrected", x3dTag );
				pose = m_p3d.TransformBy(myField[detection->GetId()] + x3dTag.Inverse() + m_x3CamToRobot);

				if( pose.X().value() > 0. && pose.X().value() < 54*12*0.0254 && pose.Y().value() > 0 && pose.Y().value() < 27*12*0.0254 && pose.Z().value() > 0.35 && pose.Z().value() < 0.50 ) {
					cv::Point ptCtr( int((pose.X().value() / 16.4592) * matF1.cols),  matF1.rows - int(pose.Y().value() / 8.2296 * matF1.rows) );
	//				printf( "%f -> %f -> %f\n", pose2.X().value(), pose2.X().value() / 16.4592, pose2.X().value() / 16.4592 * matF1.cols );
	//				printf( "-> %d  %d\n", ptCtr.x, ptCtr.y );
					cv::circle( matF1, ptCtr, 42, cv::Scalar( 0,255,0), 2 );
				}
				AprilTagDetection *pDet = (AprilTagDetection *)detection;
				pDet->SetCorners( grU );

				x3dTag = atPEst.Estimate(*detection);
				printHeadingData( "Corrected  ", x3dTag );
				printf( "2) %8.3f %8.3f %8.3f\n", x3dTag.X().value(), x3dTag.Y().value(), x3dTag.Z().value() );
				pDet->SetCorners( grUOrg );
#endif
				drawHeadingData( detection->GetId(), imgHeading, x3dTag );
				//pose = x3dTag.Inverse() + myField.m_grxfmTags[detection->GetId()-1];
				//pose = x3dTag.Inverse();
				pose = m_p3d.TransformBy(myField[detection->GetId()] + x3dTag.Inverse() + m_x3CamToRobot);
				Pose2d p2d = pose.ToPose2d();
//				if( pose.X().value() > 0. && pose.X().value() < 54*12*0.0254 && pose.Y().value() > 0 && pose.Y().value() < 27*12*0.0254 && pose.Z().value() > 0.35 && pose.Z().value() < 0.50 ) {
				if( pose.X().value() > 0. && pose.X().value() < 54*12*0.0254 && pose.Y().value() > 0 && pose.Y().value() < 27*12*0.0254) {
					cv::Point ptCtr( int((pose.X().value() / 16.4592) * matF1.cols),  matF1.rows - int(pose.Y().value() / 8.2296 * matF1.rows) );
	//				printf( "%f -> %f -> %f\n", pose2.X().value(), pose2.X().value() / 16.4592, pose2.X().value() / 16.4592 * matF1.cols );
	//				printf( "-> %d  %d\n", ptCtr.x, ptCtr.y );
					cv::circle( matF1, ptCtr, 44, cv::Scalar( 0,255,255), 2 );
				}


//				printf( "%d H:%d DM:%f -> %8.3f %8.3f %8.3f %6.2f\n", detection->GetId(), detection->GetHamming(), detection->GetDecisionMargin(), p2d.X().value(), p2d.Y().value(), pose.Z().value(), p2d.Rotation().Degrees().value() );

				myField.GetTagCorners( detection->GetId(), vFieldPts );
				int nRMin = 1000;
				int nRMax = 0;
				int nCMin = 1000;
				int nCMax = 0;
				vector<cv::Point2f> vTagCorners;
				for( int c = 0; c < 4; c++ ) {
					auto p = detection->GetCorner( c );
					cv::Point2f p2f( p.x, p.y );

					vCamCoords.push_back( p2f );
					vTagCorners.push_back( p2f );
					if( p.y < nRMin ) nRMin = p.y;
					if( p.y > nRMax ) nRMax = p.y;
					if( p.x < nCMin ) nCMin = p.x;
					if( p.x > nCMax ) nCMax = p.x;
				}
				nRMin -= 3;
				nRMax += 3;
				nCMin -= 3;
				nCMax += 3;
				if( nRMin < 0 ) nRMin = 0;
				if( nRMax >= img.rows ) nRMax = img.rows -1;
				if( nCMin < 0 ) nCMin = 0;
				if( nCMax >= img.cols ) nCMax = img.cols -1;
				
				Transform3d xfm = CalcXfrm( vTagCorners, vTagPts ).Inverse();
				Pose3d pose2 = m_p3d.TransformBy(myField[detection->GetId()] + xfm.Inverse() + m_x3CamToRobot);

//				printf( "x3dTag:   %8.3f %8.3f %8.3f     %6.3f  %6.3f  %6.3f\n", x3dTag.X().value(), x3dTag.Y().value(), x3dTag.Z().value(), x3dTag.Rotation().X().value(), x3dTag.Rotation().Y().value(), x3dTag.Rotation().Z().value() );
//				printf( "xfm:      %8.3f %8.3f %8.3f     %6.3f  %6.3f  %6.3f\n", xfm.X().value(), xfm.Y().value(), xfm.Z().value(), xfm.Rotation().X().value(), xfm.Rotation().Y().value(), xfm.Rotation().Z().value() );
//				printf( "pose:     %8.3f %8.3f %8.3f\n", pose.X().value(), pose.Y().value(), pose.Z().value() );
//				printf( "pose2:    %8.3f %8.3f %8.3f\n", pose2.X().value(), pose2.Y().value(), pose2.Z().value() );

				if( pose2.X().value() > 0. && pose2.X().value() < 54*12*0.0254 && pose2.Y().value() > 0 && pose2.Y().value() < 27*12*0.0254 && pose2.Z().value() > 0.35 && pose2.Z().value() < 0.50 ) {
					cv::Point ptCtr( int((pose2.X().value() / 16.4592) * matF1.cols),  matF1.rows - int(pose2.Y().value() / 8.2296 * matF1.rows) );
	//				printf( "%f -> %f -> %f\n", pose2.X().value(), pose2.X().value() / 16.4592, pose2.X().value() / 16.4592 * matF1.cols );
	//				printf( "-> %d  %d\n", ptCtr.x, ptCtr.y );
					cv::circle( matF1, ptCtr, int(27/(27.*12.) * matF1.rows), cv::Scalar( 255,0,0), 2 );
				}
				if( pose.X().value() > 0. && pose.X().value() < 54*12*0.0254 && pose.Y().value() > 0 && pose.Y().value() < 27*12*0.0254 && pose.Z().value() > 0.35 && pose.Z().value() < 0.50 ) {
					cv::Point ptCtr( int((pose.X().value() / 16.4592) * matF1.cols),  matF1.rows - int(pose.Y().value() / 8.2296 * matF1.rows) );
	//				printf( "%f -> %f -> %f\n", pose.X().value(), pose.X().value() / 16.4592, pose.X().value() / 16.4592 * matF1.cols );
	//				printf( "-> %d  %d\n", ptCtr.x, ptCtr.y );
					cv::circle( matF1, ptCtr, int(26./(27.*12.) * matF1.rows), cv::Scalar( 0,0,255), 2 );
				}
				Mat matTag( nRMax - nRMin + 1, nCMax - nCMin + 1, CV_8UC1, img.ptr() + nRMin * img.cols + nCMin, img.cols );

				double value = dCalcBlurr( matTag );
//				printf( "Laplacian = %lf\n", value );
				imshow( "Tag", matTag );
//				printf( "Tag Size = %dx%d\n", nRMax - nRMin + 1, nCMax - nCMin + 1);

				
#if 0
				// Get current time since power on
				clock_gettime(CLOCK_MONOTONIC, &tsCur );
				TIMESPEC_TO_TIMEVAL( &tvCur, &tsCur );

				// Calculate delta between when frame was catured and now.
				timersub( &tvCur, pFrm->ts(), &tvDelta );

				// Get current NetTables time and subtract off the delta
				int64_t time = nt::Now();
				time -= tvDelta.tv_usec;
				
				m_fpXt.Set( p2d.X().value(), time );
				m_fpYt.Set( p2d.Y().value(), time );
				m_fpRz.Set( p2d.Rotation().Radians().value(), time );
#endif
				for (int c = 0; c<4; c++ ) {
					auto p = detection->GetCorner( c );
					cv::Point cp1( (int)p.x, (int)p.y );

//					char szTmp[25];
//					sprintf( szTmp, "%d", c + 1 );
//					cv::putText( clr, szTmp, cp1, cv::FONT_HERSHEY_DUPLEX, 1.0, cv::Scalar(255,255,255), 2);
					if (c < 3) {
						auto p2 = detection->GetCorner( c + 1 );
						cv::Point cp2;

						cp2.x = p.x + (p2.x - p.x) * (c + 1) / 4;
						cp2.y = p.y + (p2.y - p.y) * (c + 1) / 4;
						cv::line( clr, cp1, cp2, cv::Scalar( 255,255,255 ), 1 );
					}
				}
			}
#if INCLUDE_GUI
			mvprintw( nDL++, 1, "Have detection: %d", detection->GetId() );
			clrtoeol();
			mvprintw( nDL++, 1, "%s", detection->GetFamily().data() );
			clrtoeol();
			mvprintw( nDL, 30, "%8.2f %8.2f %8.2f", pose.X(), pose.Y(), pose.Z() );
			mvprintw( nDL + 1, 30, "%8.4f %8.4f %8.4f", pose.Rotation().X(), pose.Rotation().Y(), pose.Rotation().Z() );
			for (int c = 0; c<4; c++ ) {
				auto p = detection->GetCorner( c );
				mvprintw( nDL++, 1, "   %.2lf  %.2lf", p.x, p.y );
			}
#endif
		}
		if( vCamCoords.size() ) {
			imshow("Heading", imgHeading );
			RobotPoseEstimator rpe( cp, vCamCoords, vFieldPts );
			auto r = rpe.EstimatePose2D();
			if( r.X().value() > 0. && r.X().value() < 54*12*0.0254 && r.Y().value() > 0 && r.Y().value() < 27*12*0.0254 ) {
				cv::Point ptCtr( int((r.X().value() / 16.4592) * matF1.cols),  matF1.rows - int(r.Y().value() / 8.2296 * matF1.rows) );
//				printf( "%f -> %f -> %f\n", r.X().value(), r.X().value() / 16.4592, r.X().value() / 16.4592 * matF1.cols );
//				printf( "-> %d  %d\n", ptCtr.x, ptCtr.y );
				cv::circle( matF1, ptCtr, int(24./(27.*12.) * matF1.rows), cv::Scalar( 0,0,255), 1 );
			}

			Transform3d xfm = CalcXfrm( vCamCoords, vFieldPts ).Inverse();
//			printf( "Inv: %9.2f %9.2f %9.2f\n", xfm.X().value(), xfm.Y().value(), xfm.Z().value() );
//			cout << "inv: " << xfm.Translation() << endl;
			imshow("Camera", clr);
			if( xfm.X().value() > 0. && xfm.X().value() < 54*12*0.0254 && xfm.Y().value() > 0 && xfm.Y().value() < 27*12*0.0254 && xfm.Z().value() > 0.35 && xfm.Z().value() < 0.50 ) {
				cv::Point ptCtr( int((xfm.X().value() / 16.4592) * matF1.cols),  matF1.rows - int(xfm.Y().value() / 8.2296 * matF1.rows) );
//				printf( "%f -> %f -> %f\n", xfm.X().value(), xfm.X().value() / 16.4592, xfm.X().value() / 16.4592 * matF1.cols );
//				printf( "-> %d  %d\n", ptCtr.x, ptCtr.y );
				cv::circle( matF1, ptCtr, int(24./(27.*12.) * matF1.rows), cv::Scalar( 255,255,255), -1 );
			}
			imshow("Field", matF1 );
			waitKey(0);
		}
	}

	return 0;
}
